@@ -21,12 +21,14 @@ along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
 """
 
+import datetime
 import http.client
 import json
 import threading
 
 from robot.api import logger
 
+from HttpCtrl.internal_messages import IgnoreRequest
 from HttpCtrl.http_server import HttpServer
 from HttpCtrl.request_storage import RequestStorage
 from HttpCtrl.response_storage import ResponseStorage
@@ -113,6 +115,9 @@ class Client:
         self.__response_body = None
         self.__response_headers = None
 
+        self.__event_queue = threading.Condition()
+        self.__async_queue = {}
+
 
     def initialize_client(self, host, port=None):
         """
@@ -190,6 +195,24 @@ class Client:
             connection.close()
 
 
+    def __wait_response_async(self, connection):
+        try:
+            server_response = connection.getresponse()
+
+            with self.__event_queue:
+                response = Response(server_response.status, server_response.read().decode('utf-8'),
+                                    server_response.headers)
+
+                self.__async_queue[connection] = response
+                self.__event_queue.notify_all()
+
+        except Exception as exception:
+            logger.info("Server has not provided response to the request (reason: %s)." % str(exception))
+
+        finally:
+            connection.close()
+
+
     def __send_request(self, connection_type, method, url, body):
         connection = self.__send(connection_type, method, url, body)
         self.__wait_response(connection)
@@ -198,10 +221,12 @@ class Client:
     def __sent_request_async(self, connection_type, method, url, body):
         connection = self.__send(connection_type, method, url, body)
 
-        wait_thread = threading.Thread(target=self.__wait_response, args=(connection,))
+        wait_thread = threading.Thread(target=self.__wait_response_async, args=(connection,))
         wait_thread.daemon = True
         wait_thread.start()
 
+        return connection
+
 
     def send_http_request(self, method, url, body=None):
         """
@@ -244,7 +269,8 @@ class Client:
         """
 
         Send HTTP request with specified parameters asynchronously. Non-blocking function to send request that waits
-        for reply using separate thread.
+        for reply using separate thread. Return connection object that is used as a key to get asynchronous response
+        using function 'Get Async Response'.
 
         `method` [in] (string): Method that is used to send request (GET, POST, PUT, DELETE, etc., see: RFC 7231, RFC 5789).
 
@@ -254,16 +280,16 @@ class Client:
 
         Example where PUT request is sent with specific body:
 
-        +-------------------------+-----+------+---------------+
-        | Send HTTP Request Async | PUT | /put | Hello Server! |
-        +-------------------------+-----+------+---------------+
+        +----------------+-------------------------+-----+------+---------------+
+        | ${connection}= | Send HTTP Request Async | PUT | /put | Hello Server! |
+        +----------------+-------------------------+-----+------+---------------+
 
         .. code:: text
 
-            Send HTTP Request Async   PUT   /put   Hello Server!
+            ${connection}=   Send HTTP Request Async   PUT   /put   Hello Server!
 
         """
-        self.__sent_request_async('http', method, url, body)
+        return self.__sent_request_async('http', method, url, body)
 
 
     def send_https_request(self, method, url, body=None):
@@ -296,7 +322,8 @@ class Client:
         """
 
         Send HTTPS request with specified parameters asynchronously. Non-blocking function to send request that waits
-        for reply using separate thread.
+        for reply using separate thread. Return connection object that is used as a key to get asynchronous response
+        using function 'Get Async Response'.
 
         `method` [in] (string): Method that is used to send request (GET, POST, DELETE, etc., see: RFC 7231, RFC 5789).
 
@@ -306,16 +333,16 @@ class Client:
 
         Example where DELETE request is sent with specific body:
 
-        +--------------------------+--------+---------+
-        | Send HTTPS Request Async | DELETE | /delete |
-        +--------------------------+--------+---------+
+        +----------------+--------------------------+--------+---------+
+        | ${connection}= | Send HTTPS Request Async | DELETE | /delete |
+        +----------------+--------------------------+--------+---------+
 
         .. code:: text
 
-            Send HTTPS Request Async   DELETE   /delete
+            ${connection}=   Send HTTPS Request Async   DELETE   /delete
 
         """
-        self.__sent_request_async('http', method, url, body)
+        return self.__sent_request_async('http', method, url, body)
 
 
     def set_request_header(self, key, value):
@@ -417,6 +444,135 @@ class Client:
             return body
 
 
+    def get_async_response(self, connection, timeout=0):
+        """
+
+        Return response as an object for the specified connection. This method should be called once after
+        'Send HTTP Request Async' or 'Send HTTPS Request Async'. It returns None if there is no response for the
+        specified connection.
+
+        `connection` [in] (object): Connection for that response should be obtained.
+
+        `timeout` [in] (int): Period of time in seconds to obtain response (by default is 0).
+
+        Example how to get response object:
+
+        +--------------+--------------------+
+        | ${response}= | Get Async Response |
+        +--------------+--------------------+
+
+        Example how to try to get response object during 10 seconds:
+
+        +--------------+--------------------+----+
+        | ${response}= | Get Async Response | 10 |
+        +--------------+--------------------+----+
+
+        .. code:: text
+
+            ${connection}=   Send HTTP Async Request   POST            /post   Hello Server!
+            ${response}=     Get Async Response        ${connection}   5
+
+        """
+        with self.__event_queue:
+            start_time = datetime.datetime.now()
+            end_time = start_time
+
+            while (end_time - start_time).total_seconds() < int(timeout):
+                if connection in self.__async_queue:
+                    return self.__async_queue.pop(connection)
+
+                self.__event_queue.wait_for(connection not in self.__async_queue, int(timeout))
+
+            return self.__async_queue.pop(connection, None)
+
+
+    def get_status_from_response(self, response):
+        """
+
+        Return response status as an integer value from the specified response object that was obtained by function
+        'Get Async Response'. Return 'None' if response object is None.
+
+        Example how to get response status from a response object:
+
+        +---------------------+--------------------------+-------------+
+        | ${response status}= | Get Status From Response | ${response} |
+        +---------------------+--------------------------+-------------+
+
+        .. code:: text
+
+            ${connection}=      Send HTTP Async Request   GET             /get
+
+            # Some other actions ...
+
+            ${response}=          Get Async Response         ${connection}   5
+            ${response status}=   Get Status From Response   ${response}
+
+        """
+        if response is None:
+            logger.error("Impossible to get status from 'None' response object.")
+            return None
+
+        return response.get_status()
+
+
+    def get_headers_from_response(self, response):
+        """
+
+        Return response headers as a map from the specified response object that was obtained by function
+        'Get Async Response'. Return 'None' if response object is None.
+
+        Example how to get response headers from a response object:
+
+        +----------------------+---------------------------+-------------+
+        | ${response headers}= | Get Headers From Response | ${response} |
+        +----------------------+---------------------------+-------------+
+
+        .. code:: text
+
+            ${connection}=      Send HTTP Async Request   GET             /get
+
+            # Some other actions ...
+
+            ${response}=           Get Async Response          ${connection}   5
+            ${response headers}=   Get Headers From Response   ${response}
+
+        """
+        if response is None:
+            logger.error("Impossible to get headers from 'None' response object.")
+            return None
+
+        return response.get_headers()
+
+
+    def get_body_from_response(self, response):
+        """
+
+        Return response body as a string from the specified response object that was obtained by function
+        'Get Async Response'. Return 'None' if response object is None.
+
+        Example how to get response code from a response object:
+
+        +-------------------+------------------------+-------------+
+        | ${response body}= | Get Body From Response | ${response} |
+        +-------------------+------------------------+-------------+
+
+        .. code:: text
+
+            ${connection}=      Send HTTP Async Request   GET             /get
+
+            # Some other actions ...
+
+            ${response}=        Get Async Response        ${connection}   5
+            ${response body}=   Get Body From Response    ${response}
+
+        """
+        if response is None:
+            logger.error("Impossible to get body from 'None' response object.")
+            return None
+
+        return response.get_body()
+
+
 
 class Server:
     """
@@ -569,8 +725,8 @@ class Server:
             self.__response_headers = {}
             self.__request = None
 
-            ResponseStorage.clear()
-            RequestStorage.clear()
+            ResponseStorage().clear()
+            RequestStorage().clear()
 
             self.__server = None
             self.__thread = None
@@ -597,7 +753,7 @@ class Server:
             Wait For Request
 
         """
-        self.__request = RequestStorage.pop()
+        self.__request = RequestStorage().pop()
         if self.__request is None:
             raise AssertionError("Timeout: request was not received.")
 
@@ -613,7 +769,7 @@ class Server:
         exception. Otherwise it waits for request during 'timeout' seconds. If during this time request is received then
         exception is thrown.
 
-        `timeout` [in] (float): Period of time in seconds when requests should not be received by HTTP server.
+        `timeout` [in] (int): Period of time in seconds when requests should not be received by HTTP server.
 
         Example how to wait for lack of requests.
 
@@ -636,13 +792,38 @@ class Server:
             Wait For No Request   10
 
         """
-        self.__request = RequestStorage.pop()
+        self.__request = RequestStorage().pop(int(timeout))
         if self.__request is not None:
             raise AssertionError("Request was received: %s." % self.__request)
 
         logger.info("Request is not received.")
 
 
+    def wait_and_ignore_request(self):
+        """
+
+        Command to server to wait incoming request and ignore it by closing connection. This call is blocked until HTTP
+        request arrives. Basically server receives all requests after \`Start Server\` and places them to internal
+        queue. When test call function \`Wait And Ignore Request\` it checks the queue and if it is not empty returns
+        the first request in the queue is ignore and connection is closed. If the queue is empty then function waits
+        when the server receives request and place it to the queue.
+
+        Example how to wait and ignore request.
+
+        +-------------------------+
+        | Wait And Ignore Request |
+        +-------------------------+
+
+        .. code:: text
+
+            Wait And Ignore Request
+
+        """
+        self.wait_for_request()
+        ResponseStorage().push(IgnoreRequest())
+        logger.info("Request is ignored by closing connection.")
+
+
     def get_request_method(self):
         """
 
@@ -790,7 +971,7 @@ class Server:
 
         """
         response = Response(int(status), body, self.__response_headers)
-        ResponseStorage.push(response)
+        ResponseStorage().push(response)
 
 
 
